/*
 * skel.cpp. skeleton for rdtsc benchmarks
 *
 * Copyright (C) 1999, 2001 by Manfred Spraul.
 *	All rights reserved except the rights granted by the GPL.
 *
 * Redistribution of this file is permitted under the terms of the GNU 
 * General Public License (GPL) version 2 or later.
 * $Header: /home/manfred/cvs-tree/timetest/svmsg_pong.cpp,v 1.2 2003/10/03 15:32:30 manfred Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sched.h>

// Intel recommends that a serializing instruction
// should be called before and after rdtsc.
// CPUID is a serializing instruction.
// ".align 128:" P 4 L2 cache line size
#define read_rdtsc_before(time)		\
	__asm__ __volatile__(		\
		".align 128\n\t"	\
		"xor %%eax,%%eax\n\t"	\
		"cpuid\n\t"		\
		"rdtsc\n\t"		\
		"mov %%eax,(%0)\n\t"	\
		"mov %%edx,4(%0)\n\t"	\
		"xor %%eax,%%eax\n\t"	\
		"cpuid\n\t"		\
		: /* no output */	\
		: "S"(&time)		\
		: "eax", "ebx", "ecx", "edx", "memory")

#define read_rdtsc_after(time)		\
	__asm__ __volatile__(		\
		"xor %%eax,%%eax\n\t"	\
		"cpuid\n\t"		\
		"rdtsc\n\t"		\
		"mov %%eax,(%0)\n\t"	\
		"mov %%edx,4(%0)\n\t"	\
		"xor %%eax,%%eax\n\t"	\
		"cpuid\n\t"		\
		: /* no output */	\
		: "S"(&time)		\
		: "eax", "ebx", "ecx", "edx", "memory")

#define BUILD_TESTFNC(name, text, instructions) \
void name##_dummy(void)				\
{						\
	__asm__ __volatile__(			\
		".align 4096\n\t"		\
		"xor %%eax, %%eax\n\t"		\
		: : : "eax");			\
}						\
static unsigned long name##_best = 1024*1024*1024; \
\
static void name(void) \
{ \
	unsigned long long time; \
	unsigned long long time2; \
 \
	read_rdtsc_before(time); \
	instructions; \
	read_rdtsc_after(time2); \
	if(time2-time < name##_best) { \
		printf( text ":\t%10Ld ticks; \n", \
			time2-time-zerotest_best); \
		name##_best = time2-time; \
	} \
}

void filler(void)
{
static int i = 3;
static int j;
	j = i*i;
}

#define DO_3(x) \
	do { x; x; x; } while(0)

#define DO_10(x) \
	do { x; x; x; x; x; x; x; x; x; x;} while(0)

#define DO_50(x) \
	do { DO_10(x); DO_10(x);DO_10(x); DO_10(x);DO_10(x);} while(0)


#define DO_T(y) do { \
	DO_3(filler()); \
	y; \
	DO_3(filler());} while(0)

#define DO_TEST(x) \
	do { \
		int i; \
		for(i=0;i<5000;i++) { \
			x; \
		} \
	} while(0)

//////////////////////////////////////////////////////////////////////////////

BUILD_TESTFNC(zerotest,"zerotest", DO_T((void)0));

//////////////////////////////////////////////////////////////////////////////
struct mybuf {
	long filler[7];
	long mtype;
	char buffer[1048576];
} g_buffers[2];

#define DEFAULT_LEN 512
int g_len;
int g_cpu;

int svmsg_id;

static void pong(void)
{
	// break any COW shares
	memset(g_buffers,0xfa,sizeof(g_buffers));
	g_buffers[0].mtype = 1;
	g_buffers[1].mtype = 2;
	for (;;) {
		msgrcv(svmsg_id, &g_buffers[0].mtype, g_len, 1, 0);
		msgsnd(svmsg_id, &g_buffers[1].mtype, g_len, 0);
	}
}

void init()
{
	int ret, isparent;
	svmsg_id = msgget(IPC_PRIVATE,0777|IPC_CREAT);
	if(svmsg_id == -1) {
		printf(" create failed.\n");
		exit(1);
	}
	fflush(stdout);
	isparent = fork();
	if (isparent < 0) {
		printf("fork failed!\n");
		exit(1);
	}	

	cpu_set_t d;
	if (g_cpu >= 0) {
		CPU_ZERO(&d);
		CPU_SET(g_cpu, &d);
		ret = sched_setaffinity(0, sizeof(d), &d);
		printf("%d: sched_setaffinity %d: %lxh\n",getpid(), ret, d.__bits[0]);
	}

	ret = sched_getaffinity(0, sizeof(d), &d);
	printf("%d: sched_getaffinity %d: %lxh\n",getpid(), ret, d.__bits[0]);

	if (isparent == 0) {
		pong();
	}
	// break any COW shares
	memset(g_buffers,0xfa,sizeof(g_buffers));
	g_buffers[0].mtype = 1;
	g_buffers[1].mtype = 2;

}

void p_transfer()
{
	msgsnd(svmsg_id, &g_buffers[0].mtype, g_len, 0);
	msgrcv(svmsg_id, &g_buffers[1].mtype, g_len, 2, 0);
}

//////////////////////////////////////////////////////////////////////////////

BUILD_TESTFNC(test_1, "svmsg_transfer", DO_T(p_transfer()));

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
	printf("SVMSG_PONG [len] [cpu]\n");
	if (argc > 1)
		g_len = atoi(argv[1]);
	if (g_len <= 0)
		g_len = DEFAULT_LEN;

	if (argc > 2)
		g_cpu = atoi(argv[2]);
	 else
		g_cpu = -1;

	printf("ping-pong of %d bytes, bound to cpu %d.\n", g_len, g_cpu);
	init();

	for(;;) {
		DO_TEST(zerotest());
		DO_TEST(test_1());
	}
	return 0;
}
